/*

dandelion.c (C) 2015 Claude Heiland-Allen <claude@mathr.co.uk>

Info: <http://mathr.co.uk/blog/2015-01-16_calendar_2015_dandelion.html>

License: GPL3+ <https://www.gnu.org/licenses/gpl.html>

Prerequisites:

    sudo aptitude install git libcairo2-dev libmpfr-dev lmodern
    git clone https://gitorious.org/maximus/book.git
    make -C book/code/lib

Compile:

    gcc -std=c99 -Wall -Wextra -pedantic -Ibook/code/lib \
    `pkg-config --cflags cairo` -o dandelion dandelion.c -Lbook/code/lib \
    `pkg-config --libs cairo` -lmandelbrot -lmpfr -lgmp -lm

Run:

    ./dandelion

Output:

    dandelion.png

*/

// feature test macro for M_PI
#define _BSD_SOURCE

#include <complex.h>
#include <math.h>
#include <stdbool.h>
#include <cairo.h>
#include "mandelbrot.h"

// image parameters
const int width  = 1666;
const int height = 1294;
const int border = 108;
const double radius = 323.0;
const complex double scale  = I * 40.0;
// const complex double offset = width / 2.0 + I * (radius + border);
const complex double offset = 833.0 + I * 431.0;
const double weight = 3.0;
const double fontsize = 48.0;
const int maxperiod = 8;
const int raysteps = 1024;

// map from parameter space to screen space
complex double to_screen(complex double c) { return c * scale + offset; }

// draw the dandelion
int main(int argc, char **argv) {
  (void) argc;
  (void) argv;
  // create a surface
  cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24, width, height);
  if (! surface) { return 1; }
  cairo_t *cr = cairo_create(surface);
  // clear to white
  cairo_set_source_rgb(cr, 1, 1, 1);
  cairo_paint(cr);
  // clip to circle
  complex double origin = to_screen(0.0 + I * 0.0);
  cairo_arc(cr, creal(origin), cimag(origin), radius, 0, 2 * M_PI);
  cairo_clip(cr);
  cairo_new_path(cr);
  // smooth lines
  cairo_set_source_rgb(cr, 0, 0, 0);
  cairo_set_line_cap(cr, CAIRO_LINE_CAP_ROUND);
  cairo_set_line_join(cr, CAIRO_LINE_JOIN_ROUND);
  // trace periodic rays
  mpq_t angle;
  mpq_init(angle);
  for (int period = 2; period <= maxperiod; ++period) {
    cairo_set_line_width(cr, weight / period);
    int denominator = (1 << period) - 1;
    for (int numerator = 1; numerator < denominator; ++numerator) {
      mpq_set_si(angle, numerator, denominator);
      mpq_canonicalize(angle);
      // check that the period is correct
      struct mandelbrot_binary_angle *b = mandelbrot_binary_angle_from_rational(angle);
      int bperiod = mandelbrot_binary_angle_period(b);
      mandelbrot_binary_angle_delete(b);
      if (period != bperiod) { continue; }
      // show progress output
      printf("%2d %3d/%3d\r", period, numerator, denominator);
      fflush(stdout);
      // trace the ray
      struct mandelbrot_external_ray_in_native *ray = mandelbrot_external_ray_in_native_new(angle);
      complex double endpoint = to_screen(mandelbrot_external_ray_in_native_get(ray));
      cairo_move_to(cr, creal(endpoint), cimag(endpoint));
      for (int i = 0; i < raysteps; ++i) {
        mandelbrot_external_ray_in_native_step(ray);
        endpoint = to_screen(mandelbrot_external_ray_in_native_get(ray));
        if (cairo_in_clip(cr, creal(endpoint), cimag(endpoint))) {
          cairo_line_to(cr, creal(endpoint), cimag(endpoint));
        } else {
          // cairo can glitch when drawing really far lines
          cairo_move_to(cr, creal(endpoint), cimag(endpoint));
        }
      }
      cairo_stroke(cr);
      mandelbrot_external_ray_in_native_delete(ray);
    }
  }
  mpq_clear(angle);
  // draw period 1 ray
  cairo_reset_clip(cr);
  cairo_set_line_width(cr, weight);
  complex double endpoint = to_screen(0.25 + I * 0.0);
  cairo_move_to(cr, creal(endpoint), cimag(endpoint));
  cairo_line_to(cr, creal(endpoint), height - border);
  cairo_stroke(cr);
  // add text
  cairo_select_font_face(cr, "LM Sans 10", CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);
  cairo_set_font_size(cr, fontsize);
  cairo_text_extents_t e;
  cairo_text_extents(cr, "2015", &e);
  cairo_move_to(cr, border - e.x_bearing, height - border);
  cairo_show_text(cr, "2015");
  cairo_text_extents(cr, "mathr.co.uk", &e);
  cairo_move_to(cr, width - border - e.x_bearing - e.width, height - border);
  cairo_show_text(cr, "mathr.co.uk");
  // save image
  cairo_surface_write_to_png(surface, "dandelion.png");
  cairo_destroy(cr);
  cairo_surface_destroy(surface);
  return 0;
}
